$include(constants.inc)
$include(maxq3120.inc)
;
; Protected EEPROM Operations in the MAXQ Environment
;
; This demonstrates a method of using an external EEPROM device in a way that
; protects the integrity of the data stored. The problem is that an
; interrupted write cycle can invalidate an entire page of the device, even
; if only one byte was being changed. With multi-millisecond write cycles,
; this makes the EEPROM unsuitable for many transaction-storage systems.
;
; This set of routines improves the integrity of the system in two ways:
; first, it provides a CRC for every data row in the device; and second,
; it provides a checked write mechanism so that writes are performed in
; two steps: a write and a commit. In this way, old data is not destroyed
; until new data is safely in the device.
;
; Register/Memory Usage
;
; These routines do not store any data permanently in registers or memory.
;
; Each routine will consume some registers; most notably, loop counters and
; the BP and OFFS registers. If you wish, you can save and restore these
; registers on the stack before they're destroyed, or move the registers up
; to higher accumulators. If you have a data stack defined on one of the
; data pointers, you can save the registers there. For this demonstration,
; all registers are fair game.
;
; Conventions
;
; Serial EEPROMs are byte-oriented devices; therefore, BP is operated in byte
; mode. For storing the 16-bit CRC, GR is used as the input/output register,
; and GRH and GRL can be used to get/save the high and low bytes. Same goes
; for the 16-bit page address. (Note that a 16-bit page address is used
; although only nine are needed; this means that devices up to 2 MB can be
; accommodated. Of course, these devices have an address space greater than
; 16 bits, so the low-level access routines would change as well.)
;
;=============================================================================
;=============================================================================
;=============================================================================
;=============================================================================
;
;
; SECTION 1: High-Level Routines
;
; Six high-level routines are defined: read, write, commit, rollback, check
; and cleanup.
;
;=============================================================================
;
; Read
;
; The read routine accepts a page number in A0 and a buffer address in BP.
; It performs the following steps:
;  1. Check page address for validity.
;  2. Read the page into the given buffer.
;  3. Calculate the CRC for the page.
;  4. Read the appropriate check page/item
;  5. Compare the calculated CRC to the read CRC
;  6. Return success or invalid read.
;
protected_eeprom_read:
;
; First, check page address for validity.
;
	move	a[1], a[0]					;Save page address
	move	apc, #80h					;Make sure AP points to acc0
	sub		#474						;Verify that request is below boundary
	jump	nc, per_invalid_page		;If not, it's an invalid page.
;
; Page is valid. Read page into buffer
;
	move	a[0], a[1]					;Restore page address
	sla4								;*32 = byte address
	sla
	call	eeprom_read_page			;Read that page
;
; Calculate CRC across page, return in GR
;
	call	crc_page					;Now, get the page CRC into GR
;
; Calculate which page check is on
;
	move	a[0], a[1]					;Get page address...
	move	a[1], gr					;...and save CRC
	call	calc_check_page				;Calculate check page address
	call	get_check_word				;...and get the word at that addr
	cmp		a[1]						;= to previous CRC?
	jump	ne, per_protection_fail		;If not, the page isn't valid
	move	acc, #op_successful			;Otherwise, return OK
	ret

per_protection_fail:
	move	acc, #op_protect_fail
	ret

per_invalid_page:
	move	acc, #op_invalid_page
	ret

;=============================================================================
;
; Write
;
; The write routine accepts a page address in ACC and a buffer at BP and
; writes it to the next available write buffer.
;
; It performs the following steps:
;  1. Check page address for validity.
;  2. Read all four write buffers. If anyone is occupied, fail.
;  3. Find the one buffer with expired state. Choose the next buffer.
;  4. Write data field to that buffer.
;  5. Calculate CRC over entire write buffer
;  6. Write address, CRC and state to auxilary page.
;
protected_eeprom_write:
;
; First, check page address for validity.
;
	move	a[1], a[0]					;Save page number
	move	apc, #80h					;Reset ACC system
	sub		#474						;Valid page number?
	jump	nc, pew_invalid_page
;
; Read all four status words. One word should be EXPIRED and the other three
; should be AVAILABLE. Any other combination is invalid.
;
	move	a[5], #0ffh					;A5 holds the "expired" buf address
	move	a[4], #0					;A4 is the buffer pointer

pew_search_bufs:
	move	acc, a[4]					;Which buffer are we inspecting?
	call	get_status_word				;Get status for that buffer
	cmp		#val_expired				;Expired?
	jump	ne, pew_try_occupied		;If not, try occupied.
	move	a[5], a[4]					;Save number of expired buf...
	jump	pew_next_buf_chk			;...and check the next one.

pew_try_occupied:
	cmp		#val_occupied				;Is this an occupied buffer?
	jump	e, pew_write_seq_err		;If so, can't write right now!

pew_next_buf_chk:
	move	acc, a[4]					;Get buffer pointer
	add		#1							;Point to next buffer
	move	a[4], acc					;Save buffer pointer
	cmp		#4							;Last one?
	jump	ne, pew_search_bufs			;If not, loop back
;
; I now know (1) the number of the buffer that's marked EXPIRED, and (2) that
; no buffer is marked OCCUPIED. I could make additional tests, but that meets
; the minimum requirements.
;
; Note that if NO buffer is EXPIRED (for whatever reason...this shouldn't
; happen) this routine will use buffer 0.
;
	move	acc, a[5]					;Get EXPIRED buffer number
	add		#1							;Point to NEXT buffer
	and		#3							;If buffer 3 was expired, use 0
;
; ACC now contains the buffer to use. Write data to buffer.
;
	sla									;*2 (write buffers are 2 pages)
	add		#505						;Add offset to DATA page
	sla4								;*32 to make byte addr
	sla
	call	eeprom_write_page			;Write the DATA page
;
; Perform CRC over all the data that will go into the write buffer.
;
	call	crc_page					;Calculate the CRC of the page
	move	acc, a[1]					;Get the page address again
	sra4								;I want the high-order byte
	sra4
	move	offs, #0					;Store it at 0 in the buffer
	move	@bp[offs], acc
	call	crc16						;CRC high address
	move	acc, a[1]					;Get page address again
	and		#0ffh						;Now I want low order byte
	move	@bp[++offs], acc			;Store it at next address
	call	crc16						;CRC low address
	move	acc, #val_occupied			;When we're done, buffer's OCCUPIED
	move	@bp[++offs], acc			;Store status byte
	call	crc16						;...and CRC that as well
	move	lc[0], #27					;Now, we'll CRC 27 zeroes in.

pew_crc_fill:
	move	acc, #0						;Load zero,
	move	@bp[++offs], acc			;...store it...
	call	crc16						;...and CRC it.
	djnz	lc[0], pew_crc_fill			;Loop over all fill bytes
	move	@bp[++offs], grh			;GR contains CRC. Store hi byte...
	move	@bp[++offs], grl			;...and lo byte.
;
; Now the buffer can be "occupied." Calculate byte address again.
;
	move	acc, a[5]					;Get page address of EXPIRED buffer
	add		#1							;Point to NEXT buffer
	and		#3							;Mask so 4 --> 0
	sla									;Now *2 because buffers are two words
	add		#504						;Point to CONTROL page
	sla4								;Make a byte address by multiplying by
	sla									;multiplying by 32
	call	eeprom_write_page			;Finally, write CONTROL page to EEPROM.
;
; Finally, blast previous buffer.
;
	move	acc, a[5]					;Now, get EXPIRED buffer address again
	sla									;*2 because buffers are 2 words
	add		#504						;Point to CONTROL page
	sla4								;Make byte address
	sla
	move	a[5], acc					;Save byte address of EXPIRED buffer
	call	eeprom_read_page			;Get that page
	move	offs, #2					;Offset 2 is the status byte
	move	@bp[offs], #val_available	;Mark it as available
	move	acc, a[5]					;Get byte address back
	call	eeprom_write_page			;Write the page
	move	acc, #op_successful			;Say operation was OK
	ret

pew_invalid_page:
	move	acc, #op_invalid_page
	ret

pew_write_seq_err:
	move	acc, #op_write_seq
	ret

;=============================================================================
;
; Commit
;
; This routine commits a previous write. That is, it finds the write buffer
; marked "occuped," copies its data into the main array, and updates the check
; word in the main array.
;
; Here are the steps for this operation:
;
;  1. Determine that one buffer is marked as occupied.
;  2. Calculate CRC on that buffer to guarantee data integrity.
;  3. Copy data field to designated page.
;  4. Calculate data field CRC.
;  5. Read check row associated with the data row.
;  6. Update check row with newly calculated CRC.
;  7. Calculate CRC for check row and install in the last word.
;  8. Write check row back to EEPROM.
;
protected_eeprom_commit:
;
; Read all four status words. Find the ONE buffer that is occupied.
;
	move	a[1], #0					;Use A1 as a buffer pointer

pec_search_bufs:
	move	acc, a[1]					;Get the buffer pointer
	call	get_status_word				;Get the status word for that buf
	cmp		#val_occupied				;Is this occupied?
	jump	e, pec_occ_found			;If so, continue.
	move	acc, a[1]					;Get buffer number
	add		#1							;Point to next one
	move	a[1], acc
	cmp		#4							;Have we looked at all the buffers?
	jump	ne, pec_search_bufs			;If not, check next buffer
	move	acc, #op_write_seq			;Otherwise, it's a write sequence err
	ret
;
; At this point, a buffer marked as OCCUPIED has been found. If you prefer,
; add code to ensure that it is unique, or to otherwise verify that the whole
; buffer system is coherent.
;
pec_occ_found:
	move	acc, a[1]					;Get OCCUPIED buffer number
	sla									;Next four instructions point to...
	add		#505						;...DATA row in buffer.
	sla4
	sla
	call	eeprom_read_page			;Get the data row into the buffer
;
; The data row is now in the buffer. We need to CRC the data row to start,
; then we'll continue the CRC with the control page.
;
	call	crc_page					;Calculate CRC over page
	move	a[5], gr					;Save the CRC value
;
; Now we've calculated the CRC of the data row. We need to CRC in the contents
; of the control page. Just one problem...we only have one buffer, and the
; data row is in there! So, we'll read the control row out, byte by byte, and
; sum it in.
;
	move	acc, a[1]
	sla4								;Multiply write buf number by 32
	sla2
	add		#16128						;Offset of first status word
	call	eeprom_set_address			;Set the address, leave in WRITE mode
	call	eeprom_configure_read		;Configure EEPROM for reading
	move	lc[0], #30

pec_crc_control:
	call	eeprom_read_byte			;Read one byte from EEPROM
	move	gr, a[5]
	call	crc16
	move	a[5], gr
	call	eeprom_tx_ack				;Acknowledge the byte
	djnz	lc[0], pec_crc_control
	call	eeprom_read_byte			;Read one byte from EEPROM
	move	grh, acc
	call	eeprom_tx_ack				;Acknowledge the byte
	call	eeprom_read_byte			;Read one byte from EEPROM
	move	grl, acc
	call	eeprom_rx_ack				;This effectively sends a NAK
	call	eeprom_stop					;Send stop condition
	move    a[0], gr					;Return address
	move	ap, #0						;Point to primary accumulator
	cmp		a[5]						;Is newly calculated CRC = stored val?
	jump	e, pec_ok_to_copy			;If so, copy to data array.
	move	acc, #op_data_corrupt		;Otherwise, data is corrupt. Say so.
	ret
;
; If execution gets here, the data in the buffer has been verified. Ready to
; copy to the data in the buffer to the appropriate data row.
;
pec_ok_to_copy:
	call	get_buffer_address			;Row address is still in A1
	move	a[3], acc					;Save the row address
	sla4								;Turn row address into...
	sla									;...a byte address...
	call	eeprom_write_page			;...and write the page.
	move	acc, a[3]					;Now, get row address back
	call	calc_check_page				;and find out where the check page is
	move	a[5], acc					;Put check word address in A5
	call	crc_page					;CRC the data page
	move	acc, a[5]					;Get back the check word addr
	and		#0ffe0h						;Isolate the row address
	move	a[4], gr					;Save the CRC while page is read
	call	eeprom_read_page			;Read the check row page
	move	acc, a[5]					;Get check word address again
	and		#1fh						;Now get offset on page...
	move	offs, acc					;...and load into offset reg
	move	gr, a[4]					;Get CRC back
	move	@bp[offs], grh				;And save at offsetted location
	move	@bp[++offs], grl
;
; This is an entry point for CLEANUP. If a check row is corrupt, this will
; CRC the contents of the first 30 bytes of the buffer, then update the last
; two bytes with the calculated CRC and write the buffer.
;
pec_update_chk_row:
	move	lc[0], #30					;CRC 30 bytes in buf

pec_crc_checkfld:
	call	crc_partial_page
	move	@bp[offs], grh				;CRC is in GR...store in last slot
	move	@bp[++offs], grl
	move	acc, a[5]					;Get check address again
	and		#0ffe0h						;Make ia a page address
	call	eeprom_write_page			;Write CRC row
;
; The data array and the check array have been updated. Just one thing left to
; do...update the write buffer status to EXPIRED.
;
	move	acc, a[1]					;Get buffer number
	sla									;Next four instructions...
	add		#504						;...turn the buffer number...
	sla4								;...into a CONTROL page addr.
	sla
	move	a[5], acc					;Save control page address
	call	eeprom_read_page			;Read the control page
	move	offs, #2					;Offset 2 is the status byte
	move	@bp[offs], #val_expired		;Update that location to EXPIRED
	move	acc, a[5]					;Get address back
	call	eeprom_write_page			;...and write the page
	move	acc, #op_successful			;Say we did everything okay.
	ret

;=============================================================================
;
; Rollback
;
; This routine marks the currently occupied write buffer as expired.
;
; That's it!
;
protected_eeprom_rollback:
;
; Read all four status words. Find the ONE buffer that is occupied.
;
	move	a[1], #0					;Start with write buffer 0

pel_search_bufs:
	move	acc, a[1]					;Get the buf number we're checking
	call	get_status_word				;Get the status word from the buffer
	cmp		#val_occupied				;Is this the occupied buffer?
	jump	e, pel_occ_found			;If so, jump.
	move	acc, a[1]					;Otherwise...
	add		#1							;...increment...
	move	a[1], acc					;...the buffer number.
	cmp		#4							;Are we done?
	jump	ne, pel_search_bufs			;If not, try next buffer.
	move	acc, #op_write_seq			;If no occupied, write seq err
	ret
;
; If execution comes here, an occupied buffer is found. Mark it expired.
;
pel_occ_found:
	move	acc, a[1]					;Get the occupied buffer number
	sla									;Make buffer number a...
	add		#504						;...byte address for the...
	sla4								;...control word in the...
	sla									;...EEPROM array
	move	a[5], acc					;save this address
	call	eeprom_read_page			;Read the page into the buffer
	move	offs, #2					;Status word is at offset=2
	move	@bp[offs], #val_expired		;Flag expired status
	move	acc, a[5]					;Restore byte address in EEPROM
	call	eeprom_write_page			;Write updated page
	move	acc, #op_successful			;...and say you did it okay.
	ret

;=============================================================================
;
; Check
;
; Check the storage subsystem for health. If unhealthy, return a code that can
; help CLEANUP figure out what to do to make it all better.
;
; Here's the scoop:
;
;  1. Read the four write buffer states. If they're all undefined, return the
;     code for uninitialized eeprom.
;  2. If only one buffer appears undefined, it's probably because the update
;     to the write buffer failed. Return the code for interrupted write. Note,
;     however, that the problem might have been a status update during a
;     commit operation that failed. In either case, the fix will be to mark
;     the buffer as expired.
;  3. If no buffer is marked as occupied, then the system checks ok.
;  4. If a buffer is marked as occupied and its CRC doesn't check, that's an
;     interrupted write.
;  5. If execution hits this point, then a buffer is marked as occupied and
;     its CRC checks -- which means that a data array write may have been
;     interrupted. Verify all check rows, and if one is bad, flag a protection
;     failure.
;  6. If all check rows are okay, then check every data row. If a data row is
;     bad, flag an interrupted commit.
;  7. If all data rows check, then the array is good but a write remains
;     uncommitted. Flag an uncommitted write.
;
protected_eeprom_check:
;
; First, I want to do a "histogram" of the various states.
;
	move	a[1], #0					;A1 contains num of expired bufs
	move	a[7], #0					;A7 contains num of occupied bufs
	move	a[3], #0					;A3 contains num of available bufs
	move	a[4], #0					;A4 contains num of undefined bufs
	move	a[5], #0					;A5 is the buffer pointer

pek_search_bufs:
	move	acc, a[5]					;Get current buffer
	call	get_status_word				;What status has this buffer?
	cmp		#val_expired				;Expired?
	jump	ne, pek_try_occupied		;If not, try occupied.
	move	acc, a[1]					;Get number of expired buffers
	add		#1							;Bump number of expired bufs...
	move	a[1], acc					;...and save it back
	jump	pek_next_buf_chk			;Now, check the next buffer

pek_try_occupied:
	cmp		#val_occupied				;Buf wasn't expired...occupied?
	jump	ne, pek_try_available		;If not, try available
	move	acc, a[7]					;Get num of occupied bufs
	add		#1							;Bump the value
	move	a[7], acc					;Save back num of occupied bufs
	move	a[6], a[5]					;A6 contains the number of the occ buf
	jump	pek_next_buf_chk			;Check the next buffer

pek_try_available:
	cmp		#val_available				;Not occupied...maybe available?
	jump	e, pek_available			;If so, jump.
	move	acc, a[4]					;Otherwise, get number of undef bufs
	add		#1							;Bump the number...
	move	a[4], acc					;...and save it back.
	jump	pek_next_buf_chk			;Go to the next buffer

pek_available:
	move	acc, a[3]					;If available...
	add		#1							;...bump the number of...
	move	a[3], acc					;...available buffers

pek_next_buf_chk:
	move	acc, a[5]					;Get the buffer number we're working
	add		#1							;Point to next one
	move	a[5], acc					;Save pointer back
	cmp		#4							;Are we done?
	jump	ne, pek_search_bufs			;If not, do another
;
; At this point, the histogram is complete and is in the various accumulators.
; Now we can analyze the data and determine what condition should be reported.
;
	move	acc, a[4]					;Get number of undef bufs
	cmp		#4							;Are all the buffers undefined?
	jump	ne, pek_chk_undef			;If not, see if *any* are undefined.
	move	acc, #op_uninitialized		;All are undefined? The EEPROM's blank
	ret

pek_chk_undef:
	jump	z, pek_chk_occupied			;If none are undef are any occupied?
	move	acc, #op_int_write			;An undef buf means write was broken
	ret

pek_chk_occupied:
	move	acc, a[7]					;Get number of occ bufs
	jump	nz, pek_chk_wbuf_crc		;If a buffer's occupied, check CRC
	move	acc, #op_successful			;No undef + no occup = OK buf system
	ret
;
; We've determined that there is at least one buffer that is occupied. That
; means that a write or a commit operation has been interrupted. We need to
; figure out what to tell CLEANUP that will help it fix the system.
;
pek_chk_wbuf_crc:
	move	acc, a[6]					;A6 contains the number of the occ buf
	sla4								;Convert A6 to...
	sla2								;...the byte address of...
	add		#16160						;...the occupied buffer.
	call	eeprom_read_page			;Read the DATA page from the buffer
	call	crc_page					;Calculate CRC over page
	move	acc, a[6]					;Once again get buffer number
	sla4								;Now convert it...
	sla2								;...to point to the...
	add		#16128						;...CONTROL page of the buffer.
	move	a[7], gr					;Save CRC (read_page blasts GR)
	call	eeprom_read_page			;Get the control page
	move	offs, #0					;CRC from beginning of page
	move	lc[0], #30					;...but only CRC 30 bytes.
	move	gr, a[7]
	call	crc_page_loop				;CRC the first 30 bytes
	move	acc, gr						;Save CRC in accumulator
	move	grh, @bp[offs++]			;Now get hi CRC from buffer
	move	grl, @bp[offs]				;...and lo CRC as well
	cmp		gr							;Does the CRC match?
	jump	e, pek_occ_ok				;If so, continue
	move	acc, #op_int_write			;Otherwise, WRITE probably interrupted
	ret
;
; An OCCUPIED write buffer has been found with a valid check word. That
; narrows the possibilities. Either an uncommitted write is pending, or a
; COMMIT operation was interrupted. If COMMIT was interrupted, we may have a
; corrupt data row or a bad check row. First, see if we have a bad check row.
;
pek_occ_ok:
	move	lc[1], #31					;There are 31 check rows

pek_nxt_row:
	move	acc, lc[1]					;Get number of row we're checking
	add		#472						;Add row offset
	sla4								;Convert absolute row number to...
	sla									;...a byte address
	move	a[7], lc[1]					;Save row counter (EEPROM uses LC1)
	call	eeprom_read_page			;Read check row into buffer
	move	lc[0], #30					;Now perform CRC over 15 word entries
	call	crc_partial_page			;Perform CRC over 30 bytes of buffer
	move	acc, gr						;Save off the CRC
	move	grh, @bp[offs++]			;Now get the high-order stored CRC
	move	grl, @bp[offs]				;...and the low-order
	cmp		gr							;Does CRC check?
	jump	e, pek_ckrow_ok				;If so, this check row is okay.
	move	acc, #op_protect_fail		;Otherwise, flag a protection failure
	ret

pek_ckrow_ok:
	move	lc[1], a[7]					;Restore LC1
	djnz	lc[1], pek_nxt_row			;...and if nonzero, loop for more rows
;
; At this point, every check row has been verified to have a good CRC. now we
; have to do some heavy lifting. *Every* data row must be read, its CRC
; calculated, and that CRC compared to what is stored in a check page.
;
; Let's get started.
;
	move	lc[1], #473					;There are 473 data rows.

pek_drow_check:
	move	acc, lc[1]					;Get the (excess-1) row number
	sub		#1							;Make the number zero-based
	sla4								;Turn row number into a...
	sla									;...byte address
	move	a[7], lc[1]					;...and save the page number
	call	eeprom_read_page			;Read the data row into buffer
	call	crc_page
	move	a[5], gr					;Save CRC in A5
	move	acc, lc[1]					;Get excess-1 row number
	sub		#1							;Convert to zero-based number
	call	calc_check_page				;Return address of check word
	call	get_check_word				;Now retrieve the check word
	cmp		a[5]						;Equal to calculated check word?
	jump	e, pek_drow_ok				;If okay, jump.
	move	acc, #op_int_commit			;Bad data row = int commit operation
	ret

pek_drow_ok:
	move	lc[1], a[7]					;Restore row counter
	djnz	lc[1], pek_drow_check		;If more rows, jump
;
; At this point, all rows have been checked, and none had a problem. That
; means that (1) the write buffer system has a verified, occupied write
; buffer, and (2) the data and check row systems are consistent. That means
; a write was successfully performed without being committed.
;
	move	acc, #op_uncommit_write
	ret

;=============================================================================
;
; Cleanup
;
; It's possible for an external EEPROM to find itself firmly wrapped around
; the axle. The CLEANUP routine tries to make things better with a minimum of
; ancillary data loss. Here's how it works:
;
; 1. Call CHECK. It will return a code indicating the operational state of the
;    system.
;
; 2. Inspect the code from CHECK. If it indicates an unitialized EEPROM, then
;    the fix is painfully easy: initialize the EEPROM and exit.
;
; 3. If CHECK indicates that a WRITE operation was interrupted, then one of
;    the buffers is hosed. Again, the fix is easy: just reinitialize the write
;    buffer array.
;
; 4. If CHECK indicates that a COMMIT operation was interrupted, then the
;    possibilities are that (1) write to a data row failed, or (2) the CPU
;    was interrupted between the data row write and the CRC update. In either
;    case the fix is easy: perform a COMMIT operation. Note that ROLLBACK is
;    not valid here: the data row is already corrupt, although the WRITE
;    buffer is okay.
;
; 5. If CHECK indicates that a protection failure has occurred, then a write
;    to the check table was interrupted, and every row associated with the
;    given check row must be CRC'ed and that check row updated.
;
; 6. Otherwise, the indication is either OK or UNCOMMITTED WRITE. In either of
;    these cases, CLEANUP really can't do anything. If OK, well, the system's
;    okay. If UNCOMMITTED WRITE, it's up to the application whether the write
;    should be committed or rolled back.
;
protected_eeprom_cleanup:
	call	protected_eeprom_check		;Perform a check operation.
	cmp		#op_uninitialized			;Is EEPROM initialized?
	jump	ne, pep_check_int_write		;If so, see if write was interrupted
;
; This is the initialization routine for the protected EEPROM. Here's what to
; do:
;
; 1. Clear every row of the EEPROM
;
; 2. Update the status of the first write buffer to EXPIRED, and of every
;    other write buffer to AVAILABLE.
;
	move	offs, #0ffh					;Pre-decrement offset register
	move	lc[0], #32					;32 bytes in buffer

pep_clear_buffer:
	move	@bp[++offs], #0				;Clear a byte of the buffer
	djnz	lc[0], pep_clear_buffer		;Loop through entire buffer
	move	lc[0], #512					;512 pages in entire EEPROM

pep_clear_eeprom:
	move	acc, lc[0]					;Get (excess-1) row number
	sub		#1							;Make zero-based
	sla4								;Convert row number to...
	sla									;...byte address
	call	eeprom_write_page			;Write the (zero) page to EEPROM
	djnz	lc[0], pep_clear_eeprom		;If not done, loop
;
; I'll provide an entry point here for initializing the write buffer array.
;
pep_init_write_bufs:
	move	acc, #16130					;Point directly to first status byte
	call	eeprom_set_address			;Set the address
	move	acc, #val_expired			;Make this first one EXPIRED
	call	eeprom_send_byte			;Send the data byte & get ack
	call	eeprom_stop					;Send stop condition to start write
	call	eeprom_ack_wait				;Wait for EEPROM to acknowledge write
	move	ap, #0						;EEPROM subs leave AP=2
	move	lc[0], #3					;Do three more buffers

pep_init_bufs:
	add		#64							;Point to next status byte
	call	eeprom_set_address			;Start EEPROM and set address
	move	acc, #val_available			;Make all these AVAILABLE
	call	eeprom_send_byte			;Send the data byte & get ack
	call	eeprom_stop					;Send stop condition and start write
	call	eeprom_ack_wait				;Wait for EEPROM to acknowledge write
	move	ap, #0						;EEPROM subs leave AP=2
	djnz	lc[0], pep_init_bufs		;Loop until all bufs are done
;
; If you want to include error checking, you can put "jump c,..." instructions
; in the above code to this label. Most of the EEPROM routines indicate an
; error by setting the carry flag.
;
pep_init_bufs_fail:
	ret
;
; Come here if the EEPROM was NOT uninitialized. Check to see if a write was
; interrupted.
;
pep_check_int_write:
	cmp		#op_int_write
	jump	ne, pep_check_int_cmmt
;
; A WRITE operation was interrupted, and the status of one of the buffers is
; invalid. There are three possibilities: either the write to the data row was
; interrupted so the buffer CRC doesn't check, or the write to the control row
; was interrupted so the status is invalid, or the write to the *previous*
; buffer was interrupted so that I have two available buffers, one occupied
; buffer and one invalid buffer. In any event, the data rows are safe. The
; best way to deal with this is just to reinitialize the entire write buffer
; array.
;
	jump	pep_init_write_bufs
;
; Come here if we've eliminated uninitialized EEPROM and interrupted WRITE.
; Check for an interrupted COMMIT operation.
;
pep_check_int_cmmt:
	cmp		#op_int_commit
	jump	ne, pep_check_prot_fail
;
; A COMMIT operation was interrupted, and the data array is in peril.
; Fortunately, the new data is in the write buffer array and now we can
; restart the COMMIT from scratch.
;
	jump	protected_eeprom_commit
;
; We've eliminated uninitialized EEPROM, interrupted WRITE, and interrupted
; COMMIT operation. Check for a protection failure.
;
pep_check_prot_fail:
	cmp		#op_protect_fail
	jump	ne, pep_other_fail
;
; A protect failure has been reported; that is, one of the protect rows has
; an invalid internal CRC. Fixing this is a little tricky: you must find all
; the data rows associated with the check row, CRC each one of them, and put
; the new collection of CRCs into the CRC row, perform a CRC on *that*, and
; then write the whole thing. This would be easier with two buffers, but I'm
; only assuming one, so I'll use the word write routine to perform interim
; updates on the affected CRC row. When fifteen words have been written, I'll
; do a page read, perform the CRC over 30 bytes, and then write the CRC to the
; last word.
;
; We just did a CHECK. The bad check row is given in LC[1].
;
	move	acc, lc[1]					;Get the row where protection failed
	add		#472						;Turn the protect row addresss...
	sla4								;...into a...
	sla									;...byte addresss.
	move	a[5], acc					;Save byte address in A5
;
; Calculate starting page for this check row. Do this by multiplying the check
; row number (zero based) by fifteen.
;
	move	acc, lc[1]					;Get (excess 1) check row again
	sub		#1							;Make row number zero based
	move	a[1], acc					;Save zero-based row number
	sla4								;Multiply by sixteen
	sub		a[1]						;Subtract original num (leaving *15)
	sla4								;Convert the page address...
	sla									;...to a byte address
	move	a[6], acc					;Save the row address in A6
	move	lc[1], #15					;CRC 15 pages

pep_prot_loop:
	move	a[7], lc[1]					;Save the row counter
	call	eeprom_read_page			;Get one row into buffer
	call	crc_page					;Calculate the CRC over the row
;
; CRC for one row is now in GR. Write that CRC to the check row.
;
	move	acc, #15					;First, we have to...
	sub		a[7]						;...calculate the word offset...
	sla									;...into the check row.
	add		a[5]						;Add to the byte address of check row.
	move	a[4], gr					;Save GR (set address clobbers it)
	call	eeprom_set_address			;Now set the address
	move	gr, a[4]					;Restore GR
	move	acc, grh					;Move high byte...
	call	eeprom_send_byte			;...and send it, and get ack.
	move	acc, grl					;Now move low byte...
	call	eeprom_send_byte			;...and send it, and get ack.
	call	eeprom_stop					;Send stop condition, starting write
	call	eeprom_ack_wait				;Wait for write to be acknowledged
	move	ap, #0						;EEPROM routines leave AP=2
	move	acc, a[6]					;Get row address in data array
	add		#32							;Point to next row
	move	a[6], acc					;and save it.
	move	lc[1], a[7]					;Restore counter to LC1
	djnz	lc[1], pep_prot_loop		;Loop if more rows to be CRC'd.
;
; At this point, fifteen CRC values have been written to the check row.
; COMMIT already has a routine to calculate the check row CRC, so I'll
; use its entry point.
;
	jump	pec_update_chk_row
;
; Any other condition doesn't require cleanup. A second call to CHECK will now
; return UNCOMMITTED WRITE or OK.
;
pep_other_fail:
	ret

;=============================================================================
;=============================================================================
;=============================================================================
;=============================================================================
;
;
; SECTION 2: Utility Subroutines
;
; These are routines that perform utility functions for the high-level
; interface routines.
;
;=============================================================================
;
; calc_check_page
;
; This routine accepts a data page address in acc and returns the byte address
; in the EEPROM of the CRC. The address is then ready to send to the get or
; set check word routines.
;
; Page address passed in A0, byte address passed out in A0.
;
; A1, LC0 destroyed.
;
calc_check_page:
	move	lc[0], #0					;Use LC0 as a (negative) counter
ccp_loop:
	sub		#15							;15 data pages per check page
	jump	c, ccp_page_found			;If below zero, we found the page
	djnz	lc[0], ccp_loop				;Otherwise, count and loop

ccp_page_found:
	add		#15							;Add back the last to get remainder
	push	a[1]						;Preserve A1
	move	a[1], acc					;Save remainder
	move	acc, #473					;Start with offset to check array
	sub		lc[0]						;Subtract (negative) page count
	sla4								;Now make the page address...
	sla									;...into a byte address
	add		a[1]						;Finally, add the remainder...
	add		a[1]						;...twice for word offset.
	pop		a[1]						;Restore (preserved) A1
	ret

;=============================================================================
;
; crc_page
;
; This routine accepts a buffer address in BP and returns a CRC in GR.
;
; crc_partial_page
;
; Same thing, but expects LC0 to be pre-initialized to something less than a
; page.
;
; crc_page_loop
;
; Same thing, but expects offset and GR to be preinitialized.
;
crc_page:
	move	lc[0], #32					;CRC entire page

crc_partial_page:
	move	offs, #0					;Start at beginning of page
	move	gr, #0						;Clear CRC register

crc_page_loop:
	move	acc, @bp[offs++]			;Get one byte from buffer
	call	crc16						;CRC the one byte
	djnz	lc[0], crc_page_loop		;and loop until done.
	ret

;=============================================================================
;
; get_check_word
;
; This routine accepts a check word address in A[0] and returns the check word
; at that location in A[0]. Basically, this is just an EEPROM read routine that
; gets a single word.
;
get_check_word:
	call	eeprom_set_address			;Set the address, leave in WRITE mode
	call	eeprom_configure_read		;Configure EEPROM for reading
	call	eeprom_read_byte			;Read one byte from EEPROM
	move	grh, acc
	call	eeprom_tx_ack
	call	eeprom_read_byte			;Read one byte from EEPROM
	move	grl, acc
	call	eeprom_rx_ack				;This effectively sends a NAK
	call	eeprom_stop					;Send stop condition
	move    a[0], gr					;Return check word
	move	ap, #0						;Point to primary accumulator
	ret

;=============================================================================
;
; get_status_word
;
; Given a write buffer number, return the status word associated with it. The
; status word is one byte long and resides at a fixed location within each
; write buffer.
;
; Input: Write buffer number in A0
;
; Output: Status word in A0
;
get_status_word:
	sla4								;Multiply write buf number by 32
	sla2
	add		#16130						;Offset of first status word
	call	eeprom_set_address			;Set the address, leave in WRITE mode
	call	eeprom_configure_read		;Configure EEPROM for reading
	call	eeprom_read_byte			;Read one byte from EEPROM
	call	eeprom_rx_ack				;This effectively sends a NAK
	call	eeprom_stop					;Send stop condition
	move    a[0], a[2]					;Return OK indication
	move	ap, #0						;Point to primary accumulator
	ret

;=============================================================================
;
; This is the exact routine that is used in all MAXQ utility ROMs (except
; the MAXQ3120...it uses a different algorithm). It, in turn, was adapted
; from an 8051 routine that has been in use for a very long time.
;
; Function   : CRC16
;
; Description: Computes 16-bit CRC value of the input data, given a seed.
;
; Input      : GR - current CRC value (seed, set to 0 for first calculation)
;              ACC - new byte to CRC in
;
; Output     : GR - CRC result
;
; Destroys   : PSF, ACC, A[3], A[4]
;
; NOTE       : The amount of time spent in this routine is now deterministic
;              without regard to the incoming data/seed.
;
; This routine takes advantage that for some polynomials, there is a shorter,
; byte-oriented solution to the CRC problem. Thus, we can avoid the shift-and-
; conditionally-XOR routine by judicious selection of the polynomial. This
; doesn't save bytes, but it does save cycles.
;
crc16:
;
; GRL contains the previous CRC. Since the low order will be shifted out
; anyway, we can get rid of it now.
;
	xor		GRL
	move	GRL, GRH
	move	GRH, #0
;
; Calculate parity of input byte XOR low order seed.
;
	move	A[3], ACC					;store the current accumulator
	xchn								;exchange nibbles in accumulator
	xor		A[3]						;acc low nybble contains XORed bits
	move	A[4], ACC					;move current ACC (we want bits 1:0)
	sra2								;move bits 3:2 into low 2 bits of ACC
	xor		A[4]						;low 2 bit pairs now XOR'ed in ACC
	move	A[4], ACC					;move current ACC (we want bit 0)
	sra									;bit '1' now in bit 0 of ACC
	xor		A[4]						;xor low 2 bits together
	move	c, ACC.0					;c now has the result
;
; C now contains the parity bit. If set, we XOR in the polynomial. Otherwise,
; equalize the branches with NOPs.
;
	jump	nc, crc_even_parity
	move	ACC, GR						;1 cycle
	xor		#0C001h						;2 cycles
	move	GR, ACC						;1 cycle
	sjump	crc_odd_parity				;1 cycle

crc_even_parity:
	nop									;Balance out the odd parity code path
	nop									;Balance out the odd parity code path
	nop									;Balance out the odd parity code path
	nop									;Balance out the odd parity code path
	nop									;Balance out the odd parity code path

crc_odd_parity:
	move	ACC, A[3]					;get original data back again
	and		#0FFh						;make sure high byte is 0
	sla4								;shift 'dat' left 6 times
	sla2
	move	A[3], ACC					;save 'dat'
	xor		GR							;xor with 'seed'
	move	GR, ACC						;store result in seed
	move	ACC, A[3]					;restore 'dat'
	sla									;shift 'dat' left once
	xor		GR							;xor with seed
	move	GR, ACC
	ret

;=============================================================================
;
; get_buffer_address
;
; This routine returns the address field from the write buffer designated in
; A[1]. The result is returned in the ACC.
;
get_buffer_address:
	move	acc, a[1]
	sla4								;Multiply write buf number by 32
	sla2
	add		#16128						;Offset of first status word
	call	eeprom_set_address			;Set the address, leave in WRITE mode
	call	eeprom_configure_read		;Configure EEPROM for reading
	call	eeprom_read_byte			;Read one byte from EEPROM
	move	grh, acc
	call	eeprom_tx_ack				;Acknowledge the byte
	call	eeprom_read_byte			;Read one byte from EEPROM
	move	grl, acc
	call	eeprom_rx_ack				;This effectively sends a NAK
	call	eeprom_stop					;Send stop condition
	move    a[0], gr					;Return address
	move	ap, #0						;Point to primary accumulator
	ret

;=============================================================================
;
; get_buffer_crc
;
; This routine returns the CRC field from the write buffer designated in
; A[1]. The result is returned in the ACC.
;
get_buffer_crc:
	move	acc, a[1]
	sla4								;Multiply write buf number by 32
	sla2
	add		#16158						;Offset of first status word
	call	eeprom_set_address			;Set the address, leave in WRITE mode
	call	eeprom_configure_read		;Configure EEPROM for reading
	call	eeprom_read_byte			;Read one byte from EEPROM
	move	grh, acc
	call	eeprom_tx_ack				;Acknowledge the byte
	call	eeprom_read_byte			;Read one byte from EEPROM
	move	grl, acc
	call	eeprom_rx_ack				;This effectively sends a NAK
	call	eeprom_stop					;Send stop condition
	move    a[0], gr					;Return address
	move	ap, #0						;Point to primary accumulator
	ret

;=============================================================================
;=============================================================================
;=============================================================================
;=============================================================================
;
;
; SECTION 3: EEPROM Subroutines
;
; These routines are high-level and low-level interface routines for the raw
; EEPROM interface. The highest-level routines are given first.
;
;=============================================================================
;
; eeprom_read_page
;
; This routine accepts a data page address in acc and returns the contents of
; the page at BP. Note that BP is a *byte* address, and that the WBS bit in
; DPC must be *clear* on entry.
;
; I2C devices don't allow direct, random page reads. Instead, one must issue
; a page write command, followed by a page address. Then ABORT the write,
; and issue a read command. The read will occur at the current page address
; which will be the address of the aborted write.
;
; There are four bits to work with: SCL, SDA_OUT, SDA_DIR, and SDA_IN. SCL
; is always driven. SDA is driven if SDA_DIR is high, and is undriven (that
; is, in receive mode) when SDA_DIR is low. SDA_OUT is the level to which
; SDA will be driven (if the CPU is driving) and SDA_IN reflects the state
; of the pin, regardless.
;
; This routine is ALWAYS entered with SCL high, and with SDA undriven (that
; is, externally pulled high.)
;

eeprom_read_page:
	call	eeprom_set_address			;Set the address, leave in WRITE mode
	call	eeprom_configure_read		;Configure EEPROM for reading
	move	offs, #0					;This clears pointer & sets source

eer_outer_loop:
	call	eeprom_read_byte			;Read one byte from EEPROM
;
; At this point, a complete received byte should be in ACC. Save it at
; BP[OFFS] and see if we're done
;
	move	@bp[offs], acc				;Store the byte just received
	move	acc, offs					;Now, get the offset
	add		#1							;Add one to the offset
	and		#1fh						;Mask offset
	move	offs, acc					;...and save it back
	jump	z,	eer_read_complete		;Jump if we're done with 32 bytes
	call	eeprom_tx_ack				;Otherwise, acknowledge this byte
	jump	eer_outer_loop				;...and loop back.
;
; Read is complete. Terminate the I2C cycle.
; 
eer_read_complete:
	call	eeprom_rx_ack				;This effectively sends a NAK
	call	eeprom_stop					;Send stop condition
	move    a[0], #0					;Return OK indication
	move	ap, #0						;Point to primary accumulator
	ret
;
; ee_error_exit simply returns a 1 in acc to signify failure.
;
ee_error_exit:
	move	a[0], #1
	move	ap, #0
	ret

;=============================================================================
;
; eeprom_write_page
;
; Low-level routine to write one physical page of the EEPROM. Get a buffer in
; BP and an address in A[0] and perform the write.
;
eeprom_write_page:
	call	eeprom_set_address			;Set the address, leave in write mode
	move	offs, #0					;Clear offs and select source

eewp_loop:
	move	acc, @bp[offs++]			;Get write data
	call	eeprom_send_byte			;send write data
	move	acc, offs					;Get offset for test
	and		#01fh						;Are we done?
	jump	nz,eewp_loop				;If not, do next byte
	call	eeprom_stop					;Send stop signal
	call	eeprom_ack_wait
	move    a[0], #0					;Flag no error
	move	ap, #0						;Point to primary accumulator
	ret

;=============================================================================
;
; eeprom_set_address
;
; This routine is called with a byte address in the EEPROM in A0. On exit,
; a write command has been sent to the EEPROM, along with two address bytes.
; The last byte will have been acknowledged by the EEPROM, and C will contain
; the result of the last byte sent.
;
; If you wish, you can continue sending data bytes, and the EEPROM will write
; them at the next STOP event. Or, you can abort the write and start some
; other operation.
;
; Input: A0 contains byte address to send to EEPROM
;
; Output: NC if operation succeeds
;
; Destroys: GR, and AP is left pointing at A[2]. This is generally needed by
;           other low-level routines.
;
eeprom_set_address:
	move	sda_out, #0					;Drop sda to signify start
	move	sda_dir, #1					;Drive SDA
	call	qcd							;Delay a quarter cycle
	move	ap, #2						;Use A[2] for scratch
	move	acc, #EEPROM_WRITE_CMD		;Move in the write command
	scall	eeprom_send_byte			;Send the write command
	sjump	c, ee_error_exit			;If not acknowledged, EXIT
	move	gr, a[0]					;Get the address...
	move	acc, grh					;First, I need the high byte...
	call	eeprom_send_byte			;...and send it to the interface
	move	acc, grl					;Now, the low byte...
	call	eeprom_send_byte			;...send it to the interface
	ret

;=============================================================================
;
; eeprom_configure_read
;
; Do this following an eeprom_set_address operation to configure the device
; for read. This routine sends a new start condition and a read command.
;
; No input, no output, and nothing destroyed other than what
; eeprom_send_byte alters.
;
eeprom_configure_read:
;
; At this point, the last ACK_RECEIVE call has left both SCL and SDA high,
; awaiting write data. But this is a READ...so the write must be aborted.
; To abort a WRITE, just issue a new START condition.
;
	move	scl, #0						;Drop clock
	call	qcd							;Wait a full half-cycle
	move    sda_out, #1					;Drive a 1 to the bus...
	move    sda_dir, #1					;...NOW.
	call	qcd							;Wait a quarter-cycle
	move	scl, #1						;Now raise the clock
	call	qcd							;Wait a quarter-cycle
	move	sda_out, #0					;Drop data for start bit
	call	qcd							;Wait a quarter-cycle
;
; The new START condition has been sent. Now send a the read command and
; accept the data.
;
	move	acc, #EEPROM_READ_CMD		;Now load the command byte...
	call	eeprom_send_byte			;Send the read command to the device
	ret

;=============================================================================
;
; eeprom_read_byte
;
; This routine clocks in exactly one byte from the EEPROM, and returns.
; In particular, it does *not* acknowledge the byte. This is because for
; multi-byte reads, the last byte prior to the stop condition is *not*
; acknowledged. So, since this routine will be used by several functions,
; it makes more sense *not* to have this routine perform the acknowledge. 
;
eeprom_read_byte:
	move	lc[1], #8					;Now we're ready to read eight bits
;
; This loop accepts a byte from the EEPROM. On entry, both SDA and SCL are
; one (the state EEPROM_RX_ACK leaves them in).
;
; After one byte has been received, this CPU has to send an acknowledge
; pulse. TX_ACK leaves the SDA transmitter enabled, so two instructions
; here (1) pull the SDA high for one CPU clock (125 ns at 8 MHz) and
; then (2) floats SDA.
;
eer_byte_loop:
	move	scl, #0						;Drop the clock
	call	qcd							;wait 1/4 cycle
	move	sda_out, #1					;Output high for one CPU clock...
	move	sda_dir, #0					;...then float bus
	call	qcd							;Wait 1/4 cycle
	move	scl, #1						;Raise the clock...data should be valid...
	call	qcd							;wait 1/4 cycle
	move	c, sda_in					;Get the data bit
	call	qcd							;wait 1/4 cycle
	rlc									;Move the bit in
	djnz	lc[1], eer_byte_loop		;and loop while there are more bits to get
	ret

;=============================================================================
;
; QCD -- I2C quarter cycle delay
;
; At 8 MHz, the call-nop-nop-return sequence takes 500 ns. Four of these are
; good for 2 us, or a 500 kHz clock rate. But realistically, there will be
; other intervening instructions. The clock rate is likely to be much closer
; to 50 kHz than to 500.
;
qcd:
	nop
	nop
	ret

;=============================================================================
;
; eeprom_send_byte
;
; A utility routine used by both EEPROM_READ and EEPROM_WRITE. It
; sends one byte to the EEPROM, receives (or doesn't receive) ACK, and
; returns.
;
; On entry, the clock is high, and the data line is indeterminate.
; It may be high, low, or undriven. For example, if a start condition
; has just been sent, SDA will be driven low. If sending back-to-back
; bytes, SDA will be undriven. For this reason, the SDA bit won't be
; driven until we've determined the state of the first data bit.
;
; On return, the I2C bus state is: SCL is high, SDA is floated, and
; the SDA output register is left with a 1.
; 
eeprom_send_byte:
	move	lc[1], #8					;Initialize counter

eesb_loop:
	move	scl, #0						;Lower clock
	call	qcd							;Wait 1/4 cycle
	sla									;Shift bit into acc.8
	move	c, acc.8					;Get the bit
	sjump	c, eesb_send_1				;If set, send a 1
	move	sda_out, #0					;Otherwise, send a zero.
	sjump	eesb_continue				;skip 1 set
	
eesb_send_1:
	move	sda_out, #1
	
eesb_continue:
	move	sda_dir, #1					;Drive the output bit
	call	qcd							;Delay 1/4 cycle
	move	scl, #1						;Raise clock
	call	qcd							;Delay 1/2 cycle...
	call	qcd							;...to allow slave to get bit
	djnz	lc[1], eesb_loop			;If more bits, loop
	call	eeprom_rx_ack				;See if slave acknowledges...
	move	sda_out, #1					;Set to 1 to make this deterministic
	ret

;=============================================================================
;
; Receive EEPROM Ack: Receive an acknowledge from the EEPROM and
; return its value in the carry flag.
;
; This will also generate a NAK if the device isn't driving the bus.
;
; All registers preserved.
;	
eeprom_rx_ack:
	move	scl, #0						;Lower clock
	call	qcd							;Delay 1/4 cycle
	move	sda_dir, #0					;Raise data (to test for ack)
	call	qcd							;wait 1/4 cycle
	move	scl, #1						;Raise clock (freezing ack)
	call	qcd							;wait 1/4 cycle
	move	c, sda_in					;Get the bit
	call	qcd							;wait 1/4 cycle
	ret

;=============================================================================
;
; Send EEPROM Ack: Send an acknowledge to the EEPROM
;
; On exit, SDA is still driven and clock is still high. So be careful!
; Drive SDA low immediately on exit and it'll look like a start condition!
;
; All registers preserved.
;	
eeprom_tx_ack:
	move	scl, #0						;Lower clock
	call	qcd							;Delay 1/4 cycle
	move	sda_out, #0					;Set output register to zero
	move	sda_dir, #1					;Drive SDA (acknowledge bit)
	call	qcd							;wait 1/4 cycle
	move	scl, #1						;Raise clock (freezing ack)
	call	qcd							;wait 1/4 cycle
	call	qcd							;wait 1/4 cycle
	ret

;=============================================================================
;
; eeprom_stop
;
; This routine sends a STOP condition (rising data line when clock is high).
; STOP indicates that an I2C transaction is complete, and that any latched
; data can be written to the EEPROM array.
;
; On entry, SCL is high and SDA may be in any state. Most likely, though,
; it's high due to a NAK in the immediately previous cycle. That means I
; can't just pull data down at this point -- that'd look like a START
; condition. So bring clock low, lower data line, bring clock high, raise
; data line -- that's a stop condition.
; 
eeprom_stop:
	move	scl, #0						;Drop clock
	call	qcd							;...for a quarter cycle
	move	sda_out, #0					;Now drop data
	move	sda_dir, #1					;Drive SDA
	call	qcd							;...wait 1/4 cycle
	move	scl, #1						;Bring clock high
	call	qcd							;wait 1/4 cycle
	move	sda_out, #1					;Now raise SDA (stop condition)
	call	qcd							;One last delay
	move	sda_dir, #0					;Float data line (make next op safe)
	move	c, #0						;Clear carry
	ret

;=============================================================================
;
; eeprom_ack_wait
;
; Wait for the EEPROM to complete the previous operation. It works by sending
; a start condition and a write command, and then determining if the part has
; returned an ACK.
;
; I can't assume the state of SDA on entry. It may be high (that is, idle) or
; it could be low due to a previous ACK. Don't take any chances. But I know
; SCL is high.
;
; Returns with AP pointing to A[2].
;
eeprom_ack_wait:
	call	qcd							;Assume SCL high, wait two clocks...
	call	qcd							;...for data settling.
	move	scl, #0						;Pull clock line low
	call	qcd							;...wait...
	move	sda_out, #1					;Set data register to 1
	move	sda_dir, #1					;Drive 1 on port
	call	qcd							;...wait...
	move	scl, #1						;Drive clock high.
	call	qcd							;Both data and clock...
	call	qcd							;...are now driven high by master.
	move	sda_out, #0					;Drop sda to signify start
	call	qcd							;Delay a quarter cycle
	move	ap, #2						;Use A[2] for scratch
	move	acc, #EEPROM_WRITE_CMD		;Move in the write command
	call	eeprom_send_byte			;Send the write command
	jump	c, eeprom_ack_wait			;If the command was not ack'ed, jump
	call	qcd							;If it WAS acked, wait.
	call	qcd							;...wait...
	move	scl, #0						;Dropping SCL should force SDA release
	call	qcd							;...wait...
	move	sda_out, #1					;But in case, drive output...
	move	sda_dir, #1					;...to a 1 level
	call	qcd							;...wait...
	move	scl, #1						;Return clock to 1 (ready for ext cmd)
	call	qcd							;...wait...
	ret

	end
